  
"
    __________________________________________________
    @Alejandro F. Reimondo 16/07/97
    This proyect sources are property of : Alejandro F. Reimondo
    Its use is NOT legal without written rights.
    Amenabar 3476 Piso: 2 Dto: B.
    Ciudad de Buenos Aires.
    Argentina.
    Tel: (54)-1-703-5586.
    __________________________________________________
"!

ExternalBuffer subclass: #SelfDefinedSubStructure
  instanceVariableNames: 
    ' definition offset '
  classVariableNames: ''
  poolDictionaries: ''  !

SelfDefinedSubStructure subclass: #SelfDefinedOpenArray
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: '' !

DynamicLinkLibrary variableByteSubclass: #Dynalink32
  classVariableNames: ''
  poolDictionaries: '' !


!SelfDefinedSubStructure class methods !
 
arrayOf: count elements: type inside: aSDS at: start
    "Answer aCollection of elements inside aSDS at start..."

    | array offset itemSize types |
    offset := start.
    types := self basicTypes.

    count = 0 ifTrue: [ "Open Array Structure..."
        ^type isInteger | type isSymbol
        ifTrue: [
            SelfDefinedOpenArray
                definedAs: type
                inside: aSDS
                at: start
        ]
        ifFalse: [
            SelfDefinedOpenArray
                named: type
                inside: aSDS
                at: start
        ]
    ].

    type isInteger ifTrue: [ itemSize := type ]
    ifFalse: [
        type isSymbol
        ifTrue: [ itemSize := types at: type ]
        ifFalse: [ "aSDS"
            itemSize := SelfDefinedStructure sizeInBytesFor: type
        ]
    ].

    array := Array new: count.
    1 to: count do: [:i|
        type isInteger
        ifTrue: [
            array at: i put: (aSDS bytesAtOffset: offset count: type).
        ]
        ifFalse: [
            array
                at: i
                put: (aSDS privateAt: (Array with: offset  with: type with: #yourself))
        ].
        offset := offset + itemSize.
    ].
    ^array!

basicTypes
    "Private - Returns the dictionary with basic types..."

    | types |
    types := Dictionary new.
    types
        "sizes in bytes"
        at: #byte put: 1;
        at: #short put: 2;
        at: #ushort put: 2;
        at: #long put: 4;
        at: #ulong put: 4;
        at: #handle put: 4;
        yourself.
    ^types!
  
definedAs: definition inside: aSelfDefinedStructure at: offset
    "Answer an instance of the receiver with definition
    inside aSelfDefinedStructure at offset..."

    ^"(self new: (definition at: #sizeInBytes))"
      self basicNew
        definition: definition;
        contents: aSelfDefinedStructure;
        offset: offset;
        yourself!
 
named: aString inside: aSelfDefinedStructure at: offset
        "Answer an instance of the receiver named aString
        which is inside aSelfDefinedStructure at offset..."
    | definition |
    definition := SelfDefinedStructure definitions at: aString.
    ^self definedAs: definition inside: aSelfDefinedStructure at: offset! !



!SelfDefinedSubStructure methods !
 
asParameter
    "Answer the receiver in a form suitable for passing
     as a parameter to a host system procedure call."

    ^self contents asParameter!
  
at: aSymbol
    "Answer the field named aSymbol from the receiver."

    | field selector |
    field := definition at: aSymbol
        ifAbsent: [
            ^super doesNotUnderstand:
                (Message new receiver: self selector: aSymbol arguments: #())
        ].
    ^self privateAt: field.!

at: aSymbol put: anObject
    "Set the field named aSymbol to anObject."

    | field selector |
    field := definition at: aSymbol
        ifAbsent: [
            ^super doesNotUnderstand:
                (Message new receiver: self selector: aSymbol arguments: #())
        ].
    ^self privateAt: field put: anObject!
  
byteAtOffset: anInteger
    "Answer the byte at anInteger in the receiver. Offsets
     are zero relative."

    ^contents byteAtOffset: anInteger + self offset!
   
byteAtOffset: anInteger put: anObject
    "Store anObject at anInteger in the receiver.
     Offsets are zero relative."

    ^contents byteAtOffset: anInteger + self offset put: anObject!

bytesAtOffset: anInteger count: anInteger2
        "Answer a ByteArray with anInteger2 bytes copied
         from the receiver starting at anInteger.
         Offsets are zero relative."

    ^contents bytesAtOffset: anInteger + self offset count: anInteger2!

bytesAtOffset: anInteger put: aByteObject
        "Store aByteObject at anInteger in the receiver.
         Offsets are zero relative."

    ^contents bytesAtOffset: anInteger + self offset put: aByteObject!
 
contents
    "Answer the contents of the receiver as a new SelfDefinedStructure."

    | newBytes |
    newBytes := contents bytesAtOffset: self offset count: self sizeInBytes.
    ^(SelfDefinedStructure usingBytes: newBytes)
            definition: definition;
            yourself!
  
copyToAddress: anAddress
    "Copy the bytes of the receiver to
      Windows memory at anAddress."

    anAddress
        replaceFrom: 1
        to: self sizeinBytes
        with: self contents
        startingAt: 1!
   
definition
    "Returns the instance var of the receiver
     for external manipulation"
    ^definition!

definition: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    definition := anObject!
   
doesNotUnderstand: aMessage
    "Private - See if the message selector is one of the
     field names of the receiver.  If so, access that field."
    | field s |
    s := aMessage selector.
    field := definition at: s
        ifAbsent: [^super doesNotUnderstand: aMessage].
    s last = $:
        ifTrue: [self privateAt: field put: aMessage arguments first]
        ifFalse: [^self privateAt: field].!
 
fillFromAddress: anAddress
    "Fill the receiver by copying bytes from
     Windows memory at anAddress."

    contents contents
        replaceFrom: 1 + self offset
        to: contents size + self offset
        with: anAddress
        startingAt: 1!
   
longAtOffset: anInteger
    "Answer the signed double word (32 bits) at
    anInteger in the receiver. Offsets are zero relative."

    ^contents longAtOffset: anInteger + self offset!

longAtOffset: anInteger put: anObject
    "Store anObject at anInteger in the receiver.
    Offsets are zero relative."

    ^contents longAtOffset: anInteger + self offset put: anObject!
 
offset
    "Returns the instance var of the receiver
     for external manipulation"
    ^offset!

offset: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    offset := anObject!
   
printOn: aStream
    "Append the ASCII representation of the receiver to aStream."

    | aString |
    aString := SelfDefinedStructure nameOf: self.
    (aString at: 1) isVowel
        ifTrue: [aStream nextPutAll: 'an ']
        ifFalse: [aStream nextPutAll: 'a '].
    aStream nextPutAll: aString!
 
privateAt: aField
    "Private - Answer the field described by aField."

    | newField |
    newField := aField copy.
    newField at: 1 put: aField first + self offset. "adapt offset"
    ^contents privateAt: newField!
  
privateAt: aField put: anObject
    "Private - Set the field described by aField to anObject."

    | newField |
    newField := aField copy.
    newField at: 1 put: aField first + self offset. "adapt offset"
    ^contents privateAt: newField put: anObject!
 
shortAtOffset: anInteger
    "Answer a signed short integer (16 bits) at anInteger in the
    receiver.  Offsets are zero relative."

    ^contents shortAtOffset: anInteger + self offset!
 
shortAtOffset: anInteger put: anObject
    "Store anInteger at anInteger in the receiver.
    Offsets are zero relative."

    ^contents shortAtOffset: anInteger + self offset put: anObject!
  
size
    "Answer the size in bytes of the structure contents."

    ^self sizeInBytes!
   
sizeInBytes
    "Answer the size of the receiver structure in bytes."

    ^definition at: #sizeInBytes!
 
uLongAtOffset: anInteger
    "Answer the double word (32 bits) at anInteger in the receiver.
    Offsets are zero relative."

    ^contents uLongAtOffset: anInteger + self offset!
 
uLongAtOffset: anInteger put: anObject
    "Store anObject at anInteger in the receiver.
    Offsets are zero relative."

    ^contents uLongAtOffset: anInteger + self offset put: anObject!
   
uShortAtOffset: anInteger
    "Answer the word (16 bits) at anInteger in the receiver.
    Offsets are zero relative."

    ^contents uShortAtOffset: anInteger + self offset!
  
uShortAtOffset: anInteger put: anObject
    "Store a word, anObject, at anInteger in the receiver.
    Offsets are zero relative."

    ^contents uShortAtOffset: anInteger + self offset put: anObject! !



!SelfDefinedOpenArray methods !
   
at: anIndex
    "Answer the field at anIndex or named aSymbol from the receiver
    indexes are one indexed...."

    anIndex isInteger ifFalse: [ ^super at: anIndex ].
    ^self itemAt: self itemSize * (anIndex - 1) + offset!
 
itemAt: anOffset
    "Private - Answer the item at anOffset"

    | sds dict |
    definition isInteger  | definition isSymbol
    ifTrue: [
        dict := Dictionary new.
        dict
            at: #sizeInBytes put: contents sizeInBytes;
            at: #item put: (Array with: 0 with: definition with: #yourself);
            yourself.
        sds := SelfDefinedSubStructure usingBytes: contents.
        sds
            definition: dict;
            offset: anOffset.
         ^sds item
    ].

    ^SelfDefinedSubStructure
        definedAs: definition
        inside: contents
        at: anOffset!
 
itemSize
    "Private - Returns the size of an item..."

    definition isDictionary ifTrue: [ ^self sizeInBytes ].
    definition isInteger ifTrue: [ ^definition ].
    ^self class basicTypes at: definition!
   
itemType: type
    "Change the elements item type to type.
    type can be:
        aSymbol for basic types (see class method basicTypes)
        anInteger for contiguous bytes in a ByteArray
        aString for a SDS name
    "

    definition := type.! !



!Dynalink32 class methods !
 
externalStructureNamed: aName
    "Returns the structure named aName
    at an external address or nil if absent...
    * Warning *
         Memory assosiated with this structure
         must be released with #free
    This method is recomended for building SDStructures..."

    | anAddress definition |
    (SelfDefinedStructure definitions includesKey: aName)
    ifFalse: [ self initializeStructures ].
    definition := SelfDefinedStructure definitions
        at: aName ifAbsent: [^nil].
    anAddress := ExternalAddress allocateMemory: (definition at: #sizeInBytes).
    ^self structureNamed: aName atAddress: anAddress!
  
fileName
        "Answer the receiver's file name."
     ^self implementedBySubclass!
 
hasBeenUsed
    "True if the receiver library has been used"

    ^self currentOrNil notNil!
 
initializeStructures
    "Private - Initializes the self defined structures for the DLL's.
    The format has to be somewhat like:

            self
                registerType: 'StructureName'
                as: #( ( fieldName fieldType ) - - - )

        fieldType must be one of the following:
                        byte,                        short,
                        ushort,                     long,
                        ulong,                      handle,
                        word,                        dword "
!

library
    "Returns the only instance of the receiver needed
     This method must be used for proper handling
     of instances of the receiver.
    This method returns nil if the library can't be found.
    Implementation now is the same as current"

    ^self current!
 
open: aString
        "Answer an instance of the receiver which refers
         to the Dynamic Link Library named aString.
    Touched by Ale for nil return and no error generation..."
    | answer |
    #modifiedByOStyle.
    ( answer := KernelLibrary loadLibrary: aString asParameter ) = 0
        ifTrue: [ ^nil"self osError" ].
   ^self fromInteger: answer!
   
register: aTypeDef as: aStrucName
    "Private - Register a new SelfDefinedStructure with
    name aStrucName with type aTypeDef."

    SelfDefinedStructure
        define: aStrucName
        withFields: (self registerDictionaryFor: aTypeDef)!
   
registerDictionaryFor: aTypeDef
    "Private - Returns the dictionary for building
     a new SelfDefinedStructure with type aTypeDef...
    aTypeDef is an array with the specification of the
     record as
        #( name type [rawType] )
     if rawType is not present #yourself will be assumed
    aType can be:
        one of #byte #short #ushort #long #ulong #handle #word #dword #uint #boolean
        aSelfDefinedStructure name
        an array identifying a subArray structure (
            e.g:
                #(3 word) can be used for defining an array of 3 words
                #(5 'MyStructure') can be used for define an array of 5 selfDefinedSubstructures named MyStructure
                #(0 'MyStructure') for an open array of structures named MyStructure
                #(9 '*WinRectangle') for an array to nine pointers to WinRectangle structures
            )
    "
    | offset dicc types def sds itemSize count |
    offset := 0.
    dicc := Dictionary new.
    types := Dictionary new.
    types
        "sizes in bytes"
        at: #byte put: 1;
        at: #short put: 2;
        at: #ushort put: 2;
        at: #long put: 4;
        at: #ulong put: 4;
        at: #handle put: #ushort;
        "conversion table for comfortable settings..."
        at: #int put: #long;
        at: #uint put: #ulong;
        at: #word put: #ushort;
        at: #dword put: #ulong;
        yourself.

    aTypeDef do: [:each|
        def := each copy.
        def size = 2 ifTrue: [def := def ,#(yourself)].
        sds := def at: 2.
		sds = #boolean ifTrue: [
			def
				at: 2 put: #ulong;
				at: 3 put: #asBoolean.
			sds := #ulong
		].
        (#(word dword) includes: sds) ifTrue: [
            sds := types at: sds
        ].
        (sds isArray or: [ sds isMemberOf: String]) ifFalse: [
            (def at: 2) isNumber
            ifFalse: [
                [ (types at: (def at: 2)
                    ifAbsent: [^self error: 'Invalid type definition']) isSymbol ]
                whileTrue: [ def at: 2 put: (types at: (def at: 2)) ]
            ].
        ].

        sds := def at: 2.
        dicc
            at: def first
            put: (Array
                with: offset
                with: sds
                with: def last).

        count := 1.

        sds isArray ifTrue: [ "compute SDS-Array size in bytes..."
            count := sds first.
            sds := sds last.
        ].
        (sds isMemberOf: String) ifTrue: [ "compute SDS size in bytes..."
            sds := SelfDefinedStructure sizeInBytesFor: sds
        ].
        [sds isInteger ] whileFalse: [
            sds := types at: sds ifAbsent: [ ^self error: 'Invalid Type def. ', sds ]
        ].

        offset := count * sds + offset.

        (count = 0 and: [ (each == aTypeDef last) not])
        ifTrue: [ ^self error: 'An open array must be the last field defined...' ].

    ].

    ^dicc!
 
registerType: aStrucName as: aDefinitionArray
    "Private - Register a new SelfDefinedStructure with
    name aStrucName with type aDefinitionArray.
    aDefinitionArray is an array with the specification of the
     record as
        #( name type [rawType] )
     if rawType is not present #yourself will be assumed
    "
    ^self register: aDefinitionArray as: aStrucName!

structureNamed: aName
    "Returns the structure named aName or nil if absent...
    This method is recomended for building SDStructures..."

    ^self structureNamed: aName atAddress: nil "default sds creation"!

structureNamed: aName atAddress: anAddress
    "Returns the structure named aName or nil if absent...
    This method is recomended for building SDStructures..."

    (SelfDefinedStructure definitions includesKey: aName)
    ifFalse: [ self initializeStructures ].
    (SelfDefinedStructure definitions includesKey: aName)
    ifFalse: [^nil ].
    ^anAddress isNil
        ifTrue: [ SelfDefinedStructure named: aName ]
        ifFalse: [ SelfDefinedStructure named: aName atAddress: anAddress]! !



!Dynalink32 methods !

close
    "Private - Free the DLL."

    super close. "perform default action..."
    self fillFromAddress: NullHandle. "fill with null handle for non UAEs..."!

externalStructureNamed: aName
    "Returns the structure named aName
    at an external address or nil if absent...
    * Warning *
         Memory assosiated with this structure
         must be released with #free
    This method is recomended for building SDStructures..."

    ^self class externalStructureNamed: aName!
 
free
    "Private - Free the DLL Retained for compatibility."

    ^self close!
  
isNull
    "True if the receiver is Null or has an invalid handle"
    ^self = NullHandle!

isValid
    "True if the receiver has a valid handle"
    ^self isNull not!
   
release
    "Release the receiver"
    self free.
    ^super release!

structureNamed: aSDSName
    "Returns the structure named aName or nil if absent...
    This method is recomended for building SDStructures..."
    ^self class structureNamed: aSDSName!

withPointerTo: aByteObject do: aBlock
    "Evaluates aBlock (with one parameter)
     allocating aByteObject in external memory;
     then free allocated memory...
    This service is implemented for passing pointer
    references to an API. Note that you must declare
     the api entry as an ulong parameter..."

    | buffer bits result |
    bits :=
        (aByteObject isNil  | aByteObject isByteArray or: [ aByteObject isString ])
        ifTrue: [ aByteObject ]
        ifFalse: [ aByteObject contents ]. "SDS asummed"
    buffer := ExternalAddress copyToNonSmalltalkMemory: bits.
    result := aBlock value: buffer.
    buffer free"  non sense in 32 bits... unlockAndFree".
    ^result! !


! SelfDefinedStructure methods !
privateAt: aField
    "Private - Answer the field described by aField.
    Modified By Ale for SelfDefinedSubStructures support..."

    | s sds array elementSize address |

    #modifiedByOStyle.
    (sds := aField at: 2) isArray
    ifTrue: [ "an Array Structure may be built..."
        ^(SelfDefinedSubStructure
                arrayOf: sds first
                elements: sds last
                inside: self
                at: (aField at: 1)) perform: (aField at: 3)
    ].
    (sds isMemberOf: String)
    ifTrue: [
        ((Definitions includesKey: sds) not and: [ sds first = $* ])
        ifTrue: [ "its a pointer to a WinStructure or to a SDS..."
            sds := sds copyFrom: 2 to: sds size.
            address := ExternalAddress fromInteger: (self uLongAtOffset: (aField at: 1)).
            ^((Definitions includesKey: sds) "an SDS name or WinStruct name spected..."
                ifTrue: [ self class named: sds atAddress: address ]
                ifFalse: [ (Smalltalk at: sds asSymbol) atAddress: address ]
                ) perform: (aField at: 3)
        ].
        "a SelfDefinedSubStructure may be built..."
        ^(SelfDefinedSubStructure
            named: sds inside: self
            at: (aField at: 1)) perform: (aField at: 3)
    ].

    s := AtSelectors at: (aField at: 2) ifAbsent: [
        ^(self bytesAtOffset: (aField at: 1) count: (aField at: 2))
            perform: (aField at: 3)].
    ^(self perform: s with: (aField at: 1)) perform: (aField at: 3).!
privateAt: aField put: anObject
    "Private - Set the field described by aField to anObject.
    Modified By Ale for SelfDefinedSubStructures support..."
    | s sds |

    #modifiedByOStyle.
    ((sds := aField at: 2) isMemberOf: String)
    ifTrue: [ "a SelfDefinedSubStructure field must be byte replaced..."
        ^self
            perform: #bytesAtOffset:put:
            with: (aField at: 1)
            with: anObject
    ].

    s := AtPutSelectors at: (aField at: 2) ifAbsent: [
        ^self bytesAtOffset: (aField at: 1) put: anObject].
    ^self perform: (AtPutSelectors at: (aField at: 2))
        with: (aField at: 1)
        with: anObject.! !

! SelfDefinedStructure class methods !
define: aString withFields: aDictionary
        "Define a memory structure.  aString is the
        name of the structure (to be used later as the
        argument to SelfDefinedStructure named:).  aDictionary
        describes the fields of the structure.  The
        dictionary keys are the symbol names of the fields
        and the values are 3 element Arrays.  The
        first element is the byte offset of the field (zero based)
        within the structure.  The second element is a
        field type symbol (for 1, 2, and 4 byte fields) or
        an integer indicating the number of bytes in the
        field (for larger fields).  The allowable field type symbols
        are:  #byte, #short, #ushort, #long, #ulong, #handle, #address.
        The third element is a selector.  When retrieving a
        field, the value is first extracted in 'raw' form, then
        the selector specified in the third array
        element is sent to this 'raw' field value
        and the resulting object is returned for the field
        value.  For all the field types except #handle, the
        raw value is an integer and so the third
        element of the array can simply be #yourself.
        For field type #handle the raw value is a 4
        element ByteArray so the third element of the
        array will typically be #asWindowHandle,
        #asExternalHandle, #asExternalAddress, etc.
        When the field type (the second array element) is an
        integer (for larger fields), the raw value is a
        ByteArray so the third element of the array
        might be #asString."
    | d max sizes sizeInBytes |
    d := IdentityDictionary new.
    max := Association key: nil value: #(0 0).
    aDictionary associationsDo: [:a |
        d at: a key asSymbol put: a value.
        d at: (a key, (String with: $:)) asSymbol put: a value.
        ((a value at: 1) >= (max value at: 1)) ifTrue: [max := a]].
    "Compute size in bytes."

"ORIGINAL CODE....
    sizes := Dictionary new.
    sizes
        at: #byte put: 1;
        at: #short put: 2;
        at: #ushort put: 2;
        at: #long put: 4;
        at: #ulong put: 4;
        at: #handle put: 4;
        at: #address put: 4.
    sizeInBytes := ((max value at: 1) +
        (sizes at: (max value at: 2) ifAbsent: [max value at: 2])).
    sizeInBytes := sizeInBytes \\ 2 + sizeInBytes.
"
    #modifiedByOStyle.
    sizeInBytes := (max value at: 1) + (self sizeInBytesFor: (max value at: 2)).

    d at: #sizeInBytes put: sizeInBytes.
    Definitions at: aString put: d
!
isDefined: aString
   "Answer true if an instance of a Win struct named aString
    was previously defined using #define:withFields:.
    Implemented by Ale for easy dynamic implementation
     of new structures & class initializations...
    Implemented by Ale..."

    #addedByOStyle.
   ^self definitions includesKey: aString!
sizeInBytesFor: item
    "Returns the size in bytes for a structure of name item...
    Added by Ale for SelfDefinedSubStructure functionality."

    | sizes dim |
    #addedByOStyle.

    item isInteger ifTrue: [ ^item ].
    item isArray ifTrue: [
        dim := (item at: 1) max: 1.
        ^(self sizeInBytesFor: (item at: 2)) * dim
    ].

    item isSymbol
    ifTrue: [ "basic type"
        sizes := Dictionary new.
        sizes
            at: #byte put: 1;
            at: #short put: 2;
            at: #ushort put: 2;
            at: #long put: 4;
            at: #ulong put: 4;
            at: #handle put: 4;
            at: #address put: 4;
            yourself.
        ^sizes at: item
    ].

    ((Definitions includesKey: item) not and: [ item first = $* ])
    ifTrue: [ "Pointer to a structure"
        ^self sizeInBytesFor: #ulong
    ].

    ^(Definitions at: item
        ifAbsent: [
            (Smalltalk includesKey: item asSymbol)
            ifTrue: [ ^(Smalltalk at: item asSymbol) sizeInBytes ].
            ^self error: self name,'  ''',item,''' is undefined.']
        ) at: #sizeInBytes! !